// D import file generated from 'src/core/sync/rwmutex.d'
module core.sync.rwmutex;
public import core.sync.exception;
import core.sync.condition;
import core.sync.mutex;
import core.memory;
version (Posix)
{
	import core.sys.posix.pthread;
}
class ReadWriteMutex
{
	enum Policy 
	{
		PREFER_READERS,
		PREFER_WRITERS,
	}
	this(Policy policy = Policy.PREFER_WRITERS)
	{
		m_commonMutex = new Mutex;
		if (!m_commonMutex)
			throw new SyncError("Unable to initialize mutex");
		m_readerQueue = new Condition(m_commonMutex);
		if (!m_readerQueue)
			throw new SyncError("Unable to initialize mutex");
		m_writerQueue = new Condition(m_commonMutex);
		if (!m_writerQueue)
			throw new SyncError("Unable to initialize mutex");
		m_policy = policy;
		m_reader = new Reader;
		m_writer = new Writer;
	}
	@property Policy policy()
	{
		return m_policy;
	}
	@property Reader reader()
	{
		return m_reader;
	}
	@property Writer writer()
	{
		return m_writer;
	}
	class Reader : Object.Monitor
	{
		this()
		{
			m_proxy.link = this;
			this.__monitor = &m_proxy;
		}
		@trusted void lock()
		{
			synchronized(m_commonMutex) {
				++m_numQueuedReaders;
				scope(exit) --m_numQueuedReaders;
				while (shouldQueueReader)
				m_readerQueue.wait();
				++m_numActiveReaders;
			}
		}
		@trusted void unlock()
		{
			synchronized(m_commonMutex) {
				if (--m_numActiveReaders < 1)
				{
					if (m_numQueuedWriters > 0)
						m_writerQueue.notify();
				}
			}
		}
		bool tryLock()
		{
			synchronized(m_commonMutex) {
				if (shouldQueueReader)
					return false;
				++m_numActiveReaders;
				return true;
			}
		}
		bool tryLock(Duration timeout)
		{
			synchronized(m_commonMutex) {
				if (!shouldQueueReader)
				{
					++m_numActiveReaders;
					return true;
				}
				enum zero = Duration.zero();
				if (timeout <= zero)
					return false;
				++m_numQueuedReaders;
				scope(exit) --m_numQueuedReaders;
				enum maxWaitPerCall = dur!"hours"(24 * 365);
				const initialTime = MonoTime.currTime;
				m_readerQueue.wait(timeout < maxWaitPerCall ? timeout : maxWaitPerCall);
				while (shouldQueueReader)
				{
					const timeElapsed = MonoTime.currTime - initialTime;
					if (timeElapsed >= timeout)
						return false;
					auto nextWait = timeout - timeElapsed;
					m_readerQueue.wait(nextWait < maxWaitPerCall ? nextWait : maxWaitPerCall);
				}
				++m_numActiveReaders;
				return true;
			}
		}
		private 
		{
			@property bool shouldQueueReader()
			{
				if (m_numActiveWriters > 0)
					return true;
				switch (m_policy)
				{
					case Policy.PREFER_WRITERS:
					{
						return m_numQueuedWriters > 0;
					}
					case Policy.PREFER_READERS:
					{
					}
					default:
					{
						break;
					}
				}
				return false;
			}
			struct MonitorProxy
			{
				Object.Monitor link;
			}
			MonitorProxy m_proxy;
		}
	}
	class Writer : Object.Monitor
	{
		this()
		{
			m_proxy.link = this;
			this.__monitor = &m_proxy;
		}
		@trusted void lock()
		{
			synchronized(m_commonMutex) {
				++m_numQueuedWriters;
				scope(exit) --m_numQueuedWriters;
				while (shouldQueueWriter)
				m_writerQueue.wait();
				++m_numActiveWriters;
			}
		}
		@trusted void unlock()
		{
			synchronized(m_commonMutex) {
				if (--m_numActiveWriters < 1)
				{
					switch (m_policy)
					{
						default:
						{
						}
						case Policy.PREFER_READERS:
						{
							if (m_numQueuedReaders > 0)
								m_readerQueue.notifyAll();
							else if (m_numQueuedWriters > 0)
								m_writerQueue.notify();
							break;
						}
						case Policy.PREFER_WRITERS:
						{
							if (m_numQueuedWriters > 0)
								m_writerQueue.notify();
							else if (m_numQueuedReaders > 0)
								m_readerQueue.notifyAll();
						}
					}
				}
			}
		}
		bool tryLock()
		{
			synchronized(m_commonMutex) {
				if (shouldQueueWriter)
					return false;
				++m_numActiveWriters;
				return true;
			}
		}
		bool tryLock(Duration timeout)
		{
			synchronized(m_commonMutex) {
				if (!shouldQueueWriter)
				{
					++m_numActiveWriters;
					return true;
				}
				enum zero = Duration.zero();
				if (timeout <= zero)
					return false;
				++m_numQueuedWriters;
				scope(exit) --m_numQueuedWriters;
				enum maxWaitPerCall = dur!"hours"(24 * 365);
				const initialTime = MonoTime.currTime;
				m_writerQueue.wait(timeout < maxWaitPerCall ? timeout : maxWaitPerCall);
				while (shouldQueueWriter)
				{
					const timeElapsed = MonoTime.currTime - initialTime;
					if (timeElapsed >= timeout)
						return false;
					auto nextWait = timeout - timeElapsed;
					m_writerQueue.wait(nextWait < maxWaitPerCall ? nextWait : maxWaitPerCall);
				}
				++m_numActiveWriters;
				return true;
			}
		}
		private 
		{
			@property bool shouldQueueWriter()
			{
				if (m_numActiveWriters > 0 || m_numActiveReaders > 0)
					return true;
				switch (m_policy)
				{
					case Policy.PREFER_READERS:
					{
						return m_numQueuedReaders > 0;
					}
					case Policy.PREFER_WRITERS:
					{
					}
					default:
					{
						break;
					}
				}
				return false;
			}
			struct MonitorProxy
			{
				Object.Monitor link;
			}
			MonitorProxy m_proxy;
		}
	}
	private 
	{
		Policy m_policy;
		Reader m_reader;
		Writer m_writer;
		Mutex m_commonMutex;
		Condition m_readerQueue;
		Condition m_writerQueue;
		int m_numQueuedReaders;
		int m_numActiveReaders;
		int m_numQueuedWriters;
		int m_numActiveWriters;
	}
}
